@@ -1,7 +1,7 @@
-use crate::algebra::{vec2d, vec3d, Vec2d, Vec3d};
+use crate::algebra::{vec2d, vec3d, Transform, Vec2d, Vec3d};
 
 mod rng;
-pub use rng::{rng, rng_pcg32, Pcg32, Rng};
+pub use rng::{rng_pcg32, Pcg32, Rng};
 
 pub fn concentric_sample_disk(u: Vec2d) -> Vec2d {
     let u = u * 2.0 - 1.0;
@@ -49,3 +49,16 @@ pub fn uniform_sample_hemisphere(n: Vec3d, u: Vec2d) -> (Vec3d, f64) {
 
     (result, 1.0 / (2.0 * std::f64::consts::PI))
 }
+
+pub fn cosine_sample_hemisphere(n: Vec3d, u: Vec2d) -> (Vec3d, f64) {
+    let d = concentric_sample_disk(u);
+    let z = 0.0f64.max(1.0 - d.x() * d.x() - d.y() * d.y()).sqrt();
+    let pdf = z * std::f64::consts::FRAC_1_PI;
+    let d = vec3d(d.x(), d.y(), z);
+    let n0 = vec3d(0.0, 0.0, 1.0);
+    let axis = n0.cross(n).normalize();
+    let angle = n0.angle_between(n).to_degrees();
+    let t = Transform::from_rotation_axis(axis, angle);
+    let d = t.transform_vector3d(d);
+    (d.normalize(), pdf)
+}
